import os
import re
from typing import Tuple

import torch
import musicparser as mp
from io import open


class Dictionary(object):
    def __init__(self) -> None:
        self.word2idx = {}
        self.idx2word = []

    def add_word(self, word: str) -> int:
        if word not in self.word2idx:
            self.idx2word.append(word)
            self.word2idx[word] = len(self.idx2word) - 1
        return self.word2idx[word]

    def __len__(self) -> int:
        return len(self.idx2word)


class Corpus(object):
    def __init__(self, path: str, use_existing_vocab: bool, split_train_test_valid: Tuple[float, float, float]) -> None:
        self.dictionary = Dictionary()
        self.parser = mp.MusicParser(path, split_train_test_valid)

        if not use_existing_vocab or (use_existing_vocab and (
                not os.path.exists(os.path.join(path, 'train.txt')) or not os.path.exists(
                os.path.join(path, 'test.txt')) or not os.path.exists(os.path.join(path, 'valid.txt')))):
            self.parser.midi_to_text()

        self.train = self.tokenize(os.path.join(path, 'train.txt'))
        self.valid = self.tokenize(os.path.join(path, 'valid.txt'))
        self.test = self.tokenize(os.path.join(path, 'test.txt'))

    def tokenize(self, path: str) -> torch.Tensor:
        """Tokenizes a text file."""
        assert os.path.exists(path), print(f"{path} doesn't exists")
        # Add words to the dictionary
        with open(path, 'r', encoding="utf8") as f:
            for line in f:
                words = line.split(self.parser.word_splitter)
                for word in words:
                    self.dictionary.add_word(word)

        # Tokenize file content
        with open(path, 'r', encoding="utf8") as f:
            idss = []
            for line in f:
                words = line.split(self.parser.word_splitter)
                ids = []
                for word in words:
                    if word in self.dictionary.word2idx:
                        ids.append(self.dictionary.word2idx[word])
                idss.append(torch.tensor(ids).type(torch.int64))
            ids = torch.cat(idss)

        return ids


def batchify(data: torch.Tensor, bsz: int, device: str = 'cpu') -> torch.Tensor:
    # Work out how cleanly we can divide the dataset into bsz parts.
    nbatch = data.size(0) // bsz
    # Trim off any extra elements that wouldn't cleanly fit (remainders).
    data = data.narrow(0, 0, nbatch * bsz)
    # Evenly divide the data across the bsz batches.
    data = data.view(bsz, -1).t().contiguous()
    return data.to(device)
