import os
import torch
import torch.nn as nn
import argparse
import numpy as np
from matplotlib import pyplot as plt

import data
import musicgenerator as mg

device = "cuda" if torch.cuda.is_available() else "cpu"

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--lr', type=float, default=1, help='initial learning rate')
    parser.add_argument('--epochs', type=int, default=100, help='upper epoch limit')
    parser.add_argument('--batch_size', type=int, default=8, help='batch size')
    parser.add_argument('--sequence_length', type=int, default=128, help='sequence length')
    parser.add_argument('--dimension_model', type=int, default=64, help='size of word embeddings')
    parser.add_argument('--nhead', type=int, default=4, help='the number of heads in the encoder/decoder of the transformer model')
    parser.add_argument('--dropout', type=float, default=0.6, help='dropout applied to layers (0 = no dropout)')
    parser.add_argument('--nhid', type=int, default=50, help='number of hidden units per layer')
    parser.add_argument('--nlayers', type=int, default=10, help='number of layers')
    parser.add_argument('--temperature', type=float, default=1.0, help='temperature used to generate result')
    parser.add_argument('--data_path', type=str, default='./training_data/classical_music/', help='path to folder with training data')
    parser.add_argument('--model_path', type=str, default=None, help='path to saved model')
    parser.add_argument('--plot_result', type=bool, default=True, help='plot a graph of the evolution of the loss')
    args = parser.parse_args()
    ###############################################################################
    # Paramètres
    ###############################################################################
    data_path = os.path.normpath(args.data_path)
    assert os.path.exists(data_path), f"Error: The path '{data_path}' does not exist."
    learning_rate = args.lr
    batch_size = args.batch_size
    split_train_test_valid = (0.8, 0.1, 0.1)
    model_path = os.path.normpath(args.model_path) if args.model_path else None
    if args.model_path:
        assert os.path.exists(model_path), f"Error: The path '{model_path}' does not exist."
    sequence_length = args.sequence_length
    dim_model = args.dimension_model
    num_head = args.nhead
    num_layers = args.nlayers
    num_hid = args.nhid
    dropout = args.dropout
    epochs = args.epochs
    nb_log_epoch = 5
    nb_words = 300
    temperature = args.temperature

    print("device :", device)

    ###############################################################################
    # Prépare les données
    ###############################################################################
    corpus = data.Corpus(data_path, model_path is not None, split_train_test_valid)
    train_data = data.batchify(corpus.train, batch_size, device)
    test_data = data.batchify(corpus.test, batch_size, device)
    val_data = data.batchify(corpus.valid, batch_size, device)

    ###############################################################################
    # Construit le modèle
    ###############################################################################
    ntokens = len(corpus.dictionary)

    if model_path:
        print("Load model from", model_path)
        with open(model_path, 'rb') as f:
            model = torch.load(f, map_location=device)
    else:
        model = mg.MusicGenerator(vocab_size=ntokens, dim_model=dim_model, num_head=num_head, nhid=num_hid, nlayers=num_layers, dropout=dropout).to(
            device)
        criterion = nn.CrossEntropyLoss()

        ###############################################################################
        # Entraîne le modèle
        ###############################################################################
        mg.train(
            model=model,
            criterion=criterion,
            ntokens=ntokens,
            train_data=train_data,
            test_data=test_data,
            val_data=val_data,
            sequence_length=sequence_length,
            lr=learning_rate,
            epochs=epochs,
            log_interval=(len(train_data) // sequence_length) // nb_log_epoch,
            data_path=data_path,
            plot_result = args.plot_result
        )

    ###############################################################################
    # Génère une mélodie
    ###############################################################################
    if corpus.parser.min_bpm > corpus.parser.max_bpm:
        min_temp = corpus.parser.min_bpm
        corpus.parser.min_bpm = corpus.parser.max_bpm
        corpus.parser.max_bpm = min_temp
    bpm = np.random.randint(corpus.parser.min_bpm, corpus.parser.max_bpm + 1)
    ts = corpus.parser.ts
    mg.generate(
        model=model,
        ntokens=ntokens,
        corpus=corpus,
        nb_words=nb_words,
        temperature=temperature,
        device=device,
        data_path=data_path,
        bpm=bpm,
        ts=ts
    )