import math
import os.path
import time

import torch
import torch.nn as nn
import torch.nn.functional as F

import matplotlib.pyplot as plt

import data


class PositionalEncoding(nn.Module):
    def __init__(self, d_model, dropout=0.1, max_len=5000):
        super(PositionalEncoding, self).__init__()
        self.dropout = nn.Dropout(p=dropout)

        pe = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2).float() * (-math.log(10000.0) / d_model))
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0).transpose(0, 1)
        self.register_buffer('pe', pe)

    def forward(self, x):
        x = x + self.pe[:x.size(0), :]
        return self.dropout(x)


class MusicGenerator(nn.Transformer):
    def __init__(self, vocab_size: int, dim_model: int, num_head: int, nhid: int, nlayers: int, dropout: float):
        super(MusicGenerator, self).__init__(d_model=dim_model, nhead=num_head, dim_feedforward=nhid,
                                             num_encoder_layers=nlayers)
        self.model_type = 'Transformer'
        self.src_mask = None
        self.pos_encoder = PositionalEncoding(dim_model, dropout)

        self.input_emb = nn.Embedding(vocab_size, dim_model)
        self.dim_model = dim_model
        self.decoder = nn.Linear(dim_model, vocab_size)

        self.init_weights()

    def _generate_square_subsequent_mask(self, sz):
        return torch.log(torch.tril(torch.ones(sz, sz)))

    def init_weights(self):
        initrange = 0.1
        nn.init.uniform_(self.input_emb.weight, -initrange, initrange)
        nn.init.zeros_(self.decoder.bias)
        nn.init.uniform_(self.decoder.weight, -initrange, initrange)

    def forward(self, src, has_mask=True):
        if has_mask:
            device = src.device
            if self.src_mask is None or self.src_mask.size(0) != len(src):
                mask = self._generate_square_subsequent_mask(len(src)).to(device)
                self.src_mask = mask
        else:
            self.src_mask = None

        src = self.input_emb(src) * math.sqrt(self.dim_model)
        src = self.pos_encoder(src)
        output = self.encoder(src, mask=self.src_mask)
        output = self.decoder(output)
        return F.log_softmax(output, dim=-1)


def get_batch(source: torch.Tensor, i: int, sequence_length: int) -> (torch.Tensor, torch.Tensor):
    seq_len = min(sequence_length, len(source) - 1 - i)
    data = source[i:i + seq_len]
    target = source[i + 1:i + 1 + seq_len].view(-1)
    return data, target


def evaluate(model: MusicGenerator,
             criterion: nn.NLLLoss,
             data_source: torch.Tensor,
             ntokens: int,
             sequence_length: int,
             ) -> float:
    # Turn on evaluation mode which disables dropout.
    model.eval()
    total_loss = 0.
    with torch.no_grad():
        for i in range(0, data_source.size(0) - 1, sequence_length):
            data, targets = get_batch(data_source, i, sequence_length)
            output = model(data)
            output = output.view(-1, ntokens)
            total_loss += len(data) * criterion(output, targets).item()
    return total_loss / (len(data_source) - 1)


def train(model: MusicGenerator,
          criterion: nn.Module,
          ntokens: int,
          train_data: torch.Tensor,
          test_data: torch.Tensor,
          val_data: torch.Tensor,
          sequence_length: int,
          lr: float,
          epochs: int,
          log_interval: int,
          data_path: str,
          plot_result: bool
          ) -> None:
    # Loop over epochs.
    best_val_loss = None
    plot_result_loss = []
    # At any point you can hit Ctrl + C to break out of training early.
    try:
        for epoch in range(1, epochs + 1):
            epoch_start_time = time.time()
            __train(model, criterion, ntokens, train_data, sequence_length, lr, epoch, log_interval)
            val_loss = evaluate(model, criterion, val_data, ntokens, sequence_length)
            print('-' * 89)
            print('| end of epoch {:3d} | time: {:5.2f}s | valid loss {:5.2f} | '
                  'valid ppl {:8.2f}'.format(epoch, (time.time() - epoch_start_time),
                                             val_loss, math.exp(val_loss)))
            print('-' * 89)
            # Save the model if the validation loss is the best we've seen so far.
            if not best_val_loss or val_loss < best_val_loss:
                with open(os.path.join(data_path, 'model.pt'), 'wb') as f:
                    torch.save(model, f)
                best_val_loss = val_loss
            else:
                # Anneal the learning rate if no improvement has been seen in the validation dataset.
                lr /= 4.0
            if plot_result:
                plot_result_loss.append(val_loss)

        test_loss = evaluate(model, criterion, test_data, ntokens, sequence_length)
        print('=' * 89)
        print('| End of training | test loss {:5.2f} | test ppl {:8.2f}'.format(
            test_loss, math.exp(test_loss)))
        print('=' * 89)

        if plot_result:
            plt.title("évolution du taux d'erreur de validation")
            plt.plot(list(range(epochs)), plot_result_loss)
            plt.xlabel("Epoch number")
            plt.ylabel("validaiton loss")

            plt.show()

    except KeyboardInterrupt:
        print('-' * 89)
        print('Exiting from training early')


def __train(model: MusicGenerator,
            criterion: nn.Module,
            ntokens: int,
            train_data: torch.Tensor,
            sequence_length: int,
            lr: float,
            epoch: int,
            log_interval: int
            ) -> None:
    model.train()
    total_loss = 0
    start_time = time.time()

    for batch, i in enumerate(range(0, train_data.size(0) - 1, sequence_length)):
        data, targets = get_batch(train_data, i, sequence_length)
        # Starting each batch, we detach the hidden state from how it was previously produced.
        # If we didn't, the model would try backpropagating all the way to start of the dataset.
        model.zero_grad()
        output = model(data)
        output = output.view(-1, ntokens)
        loss = criterion(output, targets)
        loss.backward()

        clip = 0.25
        torch.nn.utils.clip_grad_norm_(model.parameters(), clip)
        for p in model.parameters():
            p.data.add_(p.grad, alpha=-lr)

        total_loss += loss.item()

        if batch % log_interval == 0 and batch > 0:
            cur_loss = total_loss / log_interval
            elapsed = time.time() - start_time
            print('| epoch {:3d} | {:5d}/{:5d} batches | lr {:02.2e} | ms/batch {:5.2f} | '
                  'loss {:5.2f} | ppl {:8.2f}'.format(epoch, batch, len(train_data) // sequence_length, lr,
                                                      elapsed * 1000 / log_interval, cur_loss,
                                                      math.exp(cur_loss)))
            total_loss = 0
            start_time = time.time()
        # if args.dry_run:
        #     break


def generate(model: MusicGenerator,
             ntokens: int,
             device: str,
             nb_words: int,
             temperature: float,
             corpus: data.Corpus,
             data_path: str,
             bpm: int,
             ts: str
             ) -> None:
    model.eval()
    input = torch.randint(ntokens, (1, 1), dtype=torch.long).to(device)
    text = ""
    with torch.no_grad():
        for i in range(nb_words):
            output = model(input, False)
            word_weights = output[-1].squeeze().div(temperature).exp().cpu()
            word_idx = torch.multinomial(word_weights, 1)[0]
            word_tensor = torch.Tensor([[word_idx]]).long().to(device)
            input = torch.cat([input, word_tensor], 0)

            word = corpus.dictionary.idx2word[word_idx]
            print(word + ' ', end='')
            text += word + corpus.parser.word_splitter
    corpus.parser.text_to_midi(os.path.join(data_path, 'output.mid'), text, bpm, ts)
