import os.path
import re
from typing import Tuple

import numpy as np
from music21 import converter, tempo, note, chord, stream, meter
from collections import Counter


class MusicParser:
    def __init__(self, midi_folder_path: str, split_train_test_valid: Tuple[float, float, float]):
        assert sum(split_train_test_valid) == 1.0, "ERROR : Sum of split_train_test_valid must be 1"
        self.word_splitter = '&'
        self.ts = '4/4'
        self.midi_folder = os.path.normpath(midi_folder_path)
        self.train_split = split_train_test_valid[0]
        self.test_split = split_train_test_valid[1]
        self.valid_split = split_train_test_valid[2]
        self.max_bpm = np.random.randint(40, 120)  # init with random value
        self.min_bpm = np.random.randint(120, 200)  # init with random value

    def midi_to_text(self) -> None:
        textes = []
        tss = []
        for file in os.listdir(self.midi_folder):
            if file.endswith(".mid") and file != 'output.mid':
                bpm = 0
                nb_bpm_part = 0
                print(file)
                score = converter.parse(os.path.join(self.midi_folder, file))
                # print(score.show('text'))
                tsss = []
                for part in score.parts:
                    tsss.append(part.measure(1).timeSignature)
                tss.append(Counter(tsss).most_common(1)[0][0])
                score = score.chordify()
                notes = []
                for el in score.recurse():
                    if isinstance(el, note.Note):
                        notes.append(f"({el.nameWithOctave}|{el.duration.quarterLength})")
                    if isinstance(el, note.Rest):
                        notes.append(f"(|{el.duration.quarterLength})")
                    if isinstance(el, chord.Chord):
                        notes.append(
                            f"({' '.join([f'{p.nameWithOctave}' for p in el.pitches])}|{el.duration.quarterLength})")

                    if 'TempoIndication' in el.classSet:
                        bpm += el.number
                        nb_bpm_part += 1

                bpm = bpm // nb_bpm_part
                if not self.max_bpm:
                    self.max_bpm = bpm
                if not self.min_bpm:
                    self.min_bpm = bpm

                if bpm > self.max_bpm:
                    self.max_bpm = bpm
                if bpm < self.min_bpm:
                    self.min_bpm = bpm
                textes.append(self.word_splitter.join(notes))

        ts = Counter(tss).most_common(1)[0][0]
        self.ts = f"{ts.numerator}/{ts.denominator}"
        nb_music = len(textes)
        nb_train = int(nb_music * self.train_split)
        nb_test = int(nb_music * self.test_split)
        nb_valid = int(nb_music * self.valid_split)
        with open(os.path.join(self.midi_folder, 'train.txt'), 'w') as f:
            data = textes[:nb_train] if len(textes[:nb_train]) > 0 else [textes[0]]
            for music in data:
                f.write(f"{music}{self.word_splitter}")
        with open(os.path.join(self.midi_folder, 'test.txt'), 'w') as f:
            data = textes[nb_train:nb_train + nb_test] if len(textes[nb_train:nb_train + nb_test]) > 0 else [textes[-1]]
            for music in data:
                f.write(f"{music}{self.word_splitter}")
        with open(os.path.join(self.midi_folder, 'valid.txt'), 'w') as f:
            data = textes[nb_train + nb_test:] if len(textes[nb_train + nb_test:]) > 0 else [textes[-1]]
            for music in data:
                f.write(f"{music}{self.word_splitter}")

    def text_to_midi(self, midi_file_path: str, text: str, bpm: int, ts: str) -> None:
        # Create a Stream object to hold the notes and rests
        midi_stream = stream.Stream()

        mm = tempo.MetronomeMark(number=bpm)
        midi_stream.insert(0, mm)

        ts = meter.TimeSignature(self.ts)
        midi_stream.insert(0, ts)

        # Split the sequential text into individual tokens
        tokens = text.split('&')

        # Iterate over each token in the text
        for token in tokens:
            elem = re.search(r'\((.*?)\)', token)
            if elem:
                elem = elem.group(1).split('|')
                duration = self.__strQuarterLength_to_float(elem[1])
                if len(elem[0]) > 0:
                    # If the token is a chord or a note (a chord with one note)
                    c = chord.Chord(elem[0])
                    c.duration.quarterLength = duration
                    midi_stream.append(c)
                else:
                    # If the token is a rest
                    r = note.Rest(duration)
                    midi_stream.append(r)

        # Convert the Stream to a MIDI file and write it to disk
        midi_stream.show('text')
        midi_stream.write('midi', fp=midi_file_path)

    def __strQuarterLength_to_float(self, quarterLength: str) -> float:
        res = 0.0
        try:
            res = float(quarterLength)
        except ValueError:
            res = quarterLength.split('/')
            res = int(res[0]) / int(res[1])
        return res
